from .base import BasePipeline
from .classical import ClassicalPipeline
from .logarithm_pca import LogarithmPCAPipeline

__all__ = [
    "BasePipeline",
    "ClassicalPipeline",
    "LogarithmPCAPipeline",
]



from typing import TYPE_CHECKING, Tuple

import numpy as np
from tqdm import tqdm

if TYPE_CHECKING:
    from .calculator import Calculator


def calculate_auc_triple_parameters(
    calculator: "Calculator", grid_interval: int
) -> Tuple:
    """Calculate AUC triple parameters.

    :param grid_interval: grid interval
    :return: tuple of W1, W2, WUAUC
    """
    w1_values = np.linspace(0, 1, grid_interval)
    w2_values = np.linspace(0, 1, grid_interval)
    W1, W2 = np.meshgrid(w1_values, w2_values)
    WUAUC = np.zeros_like(W1)

    for i in tqdm(range(W1.shape[0]), desc="Progress"):
        for j in range(W1.shape[1]):
            w1 = W1[i, j]
            w2 = W2[i, j]
            w3 = 1 - w1 - w2
            if w3 < 0:
                WUAUC[i, j] = np.nan
            else:
                WUAUC[i, j] = calculator.calculate_wuauc(
                    groupby="user_id",
                    weights_for_equation=[w1, w2, w3],
                    weights_for_groups=calculator.weights_for_groups,
                )
    return W1, W2, WUAUC



from functools import partialmethod
from typing import Dict, List, Optional, Union

from optuna.trial import Trial

from ..evaluation import Calculator, LogarithmPCACalculator
from .base import BaseObjective, BaseObjectiveConfig
from .construct_weights import construct_weights
from .evaluate_targets import evaluate_targets


class MutipleObjectiveConfig(BaseObjectiveConfig):
    """Configuration for handling multiple objectives in optimization.

    Attributes:
        first_order_lower_bound (float): The lower bound for the first-order objective.
        first_order_upper_bound (float): The upper bound for the first-order objective.
        max_min_scale_ratio (Optional[float]): The maximum to minimum scale ratio. None indicates no specific ratio.
        first_order_scale_upper_bound (float): The upper scale bound for the first-order objective.
        first_order_scale_lower_bound (float): The lower scale bound for the first-order objective.
        power_lower_bound (float): The lower bound for the power objective.
        power_upper_bound (float): The upper bound for the power objective.
        pca_importance_lower_bound (float): The lower bound for PCA importance.
        pca_importance_upper_bound (float): The upper bound for PCA importance.
    """

    first_order_lower_bound: float = 1e-3
    first_order_upper_bound: float = 1e6
    max_min_scale_ratio: Optional[float] = None
    first_order_scale_upper_bound: float = 1
    first_order_scale_lower_bound: float = 1
    power_lower_bound: float = -1
    power_upper_bound: float = 1
    pca_importance_lower_bound: float = 0
    pca_importance_upper_bound: float = 10


class MultipleObjective(BaseObjective):
    """
    This class provides methods to optimize the portfolio objective.
    """

    construct_weights = partialmethod(construct_weights)

    def __init__(
        self,
        calculator: Union[Calculator, LogarithmPCACalculator],
        direction: Optional[str] = None,
        formula: Optional[str] = None,
        first_order: Optional[bool] = False,
        power: Optional[bool] = True,
        dirichlet: Optional[bool] = False,
        weights_num: Optional[int] = None,
        study_name: Optional[str] = None,
        study_path: Optional[str] = None,
        save_study: Optional[bool] = True,
        first_order_lower_bound: float = 1e-3,
        first_order_upper_bound: float = 1e6,
        max_min_scale_ratio: Optional[float] = None,
        first_order_scale_upper_bound: float = 1,
        first_order_scale_lower_bound: float = 1,
        power_lower_bound: float = -1,
        power_upper_bound: float = 1,
        pca_importance_lower_bound: float = 0,
        pca_importance_upper_bound: float = 10,
        config: Optional[Dict] = None,
    ) -> None:
        """
        Initialize with direction, weights_num, formula, and dirichlet.

        Args:
            first_order_lower_bound (float, optional): Lower bound for first order value. Defaults to 1e-3.
            first_order_upper_bound (float, optional): Upper bound for first order value. Defaults to 1e6.
            power_lower_bound (float, optional): Lower bound for power value. Defaults to -1.
            power_upper_bound (float, optional): Upper bound for power value. Defaults to 1.
            pca_importance_lower_bound (float, optional): Lower bound for pca importance value. Defaults to 0.
            pca_importance_upper_bound (float, optional): Upper bound for pca importance value. Defaults to 10.
            first_order_scale_bound (Optional[float], optional): Scale bound for first order value. Defaults to None.
        """

        if config is not None:
            self.config = MutipleObjectiveConfig(**config)
        else:
            self.config = MutipleObjectiveConfig(
                direction=direction,
                formula=formula,
                first_order=first_order,
                power=power,
                dirichlet=dirichlet,
                weights_num=weights_num,
                study_name=study_name,
                study_path=study_path,
                save_study=save_study,
                first_order_lower_bound=first_order_lower_bound,
                first_order_upper_bound=first_order_upper_bound,
                max_min_scale_ratio=max_min_scale_ratio,
                first_order_scale_upper_bound=first_order_scale_upper_bound,
                first_order_scale_lower_bound=first_order_scale_lower_bound,
                power_lower_bound=power_lower_bound,
                power_upper_bound=power_upper_bound,
                pca_importance_lower_bound=pca_importance_lower_bound,
                pca_importance_upper_bound=pca_importance_upper_bound,
            )
        self.calculator = calculator
        self.direction = self.config.direction
        self.formula = self.config.formula
        self.first_order = self.config.first_order
        self.power = self.config.power
        self.dirichlet = self.config.dirichlet
        self.weights_num = self.config.weights_num
        self.study_name = self.config.study_name
        self.study_path = self.config.study_path
        self.save_study = self.config.save_study
        self.first_order_lower_bound = self.config.first_order_lower_bound
        self.first_order_upper_bound = self.config.first_order_upper_bound
        self.max_min_scale_ratio = self.config.max_min_scale_ratio
        self.first_order_scale_upper_bound = self.config.first_order_scale_upper_bound
        self.first_order_scale_lower_bound = self.config.first_order_scale_lower_bound
        self.power_lower_bound = self.config.power_lower_bound
        self.power_upper_bound = self.config.power_upper_bound
        self.pca_importance_lower_bound = self.config.pca_importance_lower_bound
        self.pca_importance_upper_bound = self.config.pca_importance_upper_bound

        self.target_columns: List[str] = []
        self.mask_columns: List[Optional[str]] = []
        self.evaluator_flags: List[str] = []
        self.groupbys: List[Optional[str]] = []
        self.hyperparameters: List[Optional[float]] = []
        self.evaluator_propertys: List[Optional[str]] = []

        if self.power_lower_bound < 0:
            self.dirichlet = False

        self._prepare_study()

    def add_evaluator(
        self,
        flag: str,
        target_column: str,
        mask_column: Optional[str] = None,
        hyperparameter: Optional[float] = None,
        evaluator_property: Optional[str] = None,
        groupby: Optional[str] = None,
    ) -> None:
        """
        Adds evaluators to the objective.

        Args:
            flag (str): Type of calculator. Expected values include ["wuauc", "portfolio", "logmse", ...].
            target_column (str): The target column for calculation.
            hyperparameter (Optional[float], optional): Hyperparameter for the calculator. Defaults to None.
            evaluator_property (Optional[str], optional): Property of the evaluator. Defaults to None.
            groupby (Optional[str], optional): Grouping criteria. Defaults to None.
        """
        self.evaluator_flags.append(flag)
        self.target_columns.append(target_column)
        if mask_column is not None:
            self.mask_columns.append(mask_column)
        else:
            self.mask_columns.append(None)
        if hyperparameter is not None:
            self.hyperparameters.append(hyperparameter)
        else:
            self.hyperparameters.append(None)
        if groupby is not None:
            self.groupbys.append(groupby)
        else:
            self.groupbys.append(None)
        if evaluator_property is not None:
            self.evaluator_propertys.append(evaluator_property)
        else:
            self.evaluator_propertys.append(None)

    def objective(
        self,
        trial: Trial,
    ) -> float:
        """
        Objective function to be optimized by optuna.

        Args:
            trial (Trial): Optuna trial instance.

        Returns:
            float: Computed objective value based on the provided trial.
        """
        weights = construct_weights(self, trial)
        targets = evaluate_targets(
            calculator=self.calculator,
            evaluator_flags=self.evaluator_flags,
            mask_columns=self.mask_columns,
            hyperparameters=self.hyperparameters,
            evaluator_propertys=self.evaluator_propertys,
            groupbys=self.groupbys,
            target_columns=self.target_columns,
            weights=weights,
        )
        local_vars = {"targets": targets, "sum": sum, "max": max, "min": min}
        result = float(eval(str(self.formula), {"__builtins__": None}, local_vars))
        if self.logger:
            self.logger.info(f"Trial {trial.number} finished with result: {result}")
            self.logger.info(f"targets: {targets}")
            self.logger.info(f"weights: {weights}")
        return result



from typing import Iterable, List, Union, cast

import matplotlib.pyplot as plt
import numpy as np

from ..evaluation.calculator import Calculator


class PortfolioPlotter:
    """PortfolioPlotter class for plotting portfolio curve."""

    def __init__(
        self,
        calculator: Calculator,
        target_column: str,
        points_num: int = 20,
        minimal_expected_return: float = 0.9,
        colors: List[str] = [
            "orange",
            "red",
            "green",
            "blue",
            "purple",
            "brown",
            "pink",
            "gray",
            "olive",
            "cyan",
        ],
    ) -> None:
        """Initialize PortfolioPlotter.

        Args:
            calculator (Calculator): Calculator instance for calculating the portfolio.
            target_column (str): target column for calculating the portfolio.
            points_num (int, optional): sample points number. Defaults to 20.
            minimal_expected_return (float, optional): minimal expected return ratio. Defaults to 0.9.
            colors (List[str], optional): colors for plotting.
        """
        self.calculator = calculator
        self.target_column = target_column
        self.points_num = points_num
        self.minimal_expected_return = minimal_expected_return
        self.colors = colors

    def _generate_points(self) -> None:
        """Generate points for plotting."""
        self.expected_returns = np.linspace(
            self.minimal_expected_return, 1, num=self.points_num, endpoint=True
        )
        self.top_ratios = [
            self.calculator.calculate_portfolio_concentration(
                target_column=self.target_column, expected_return=er
            )[1]
            for er in self.expected_returns
        ]
        self.expected_returns = self.expected_returns[::-1]
        self.top_ratios = self.top_ratios[::-1]
        self.top_ratios = [1 - ratio for ratio in self.top_ratios]
        self.expected_returns[0] = 1
        self.top_ratios[0] = 0

    def _plot_single(self, weights_for_equation: List[float], color: str) -> None:
        """Plot single portfolio curve.

        Args:
            weights_for_equation (List[float]): weights for equation
            color (str): color for plotting curve
        """
        self.calculator.get_overall_score(weights_for_equation)
        self._generate_points()
        plt.plot(self.top_ratios, self.expected_returns, color=color)
        plt.fill_between(
            self.top_ratios,
            self.expected_returns,
            y2=self.minimal_expected_return,
            color=color,
            alpha=0.1,
        )

    def plot(
        self, weights_for_equations: Union[List[float], List[List[float]]]
    ) -> None:
        """Plot portfolio curve.

        Args:
            weights_for_equations (Union[List[float], List[List[float]]]): weights for equations
        """
        plt.figure(figsize=(10, 6))
        if isinstance(weights_for_equations[0], Iterable):
            weights_for_equations = cast(List[List[float]], weights_for_equations)
            for i, weights_for_equation in enumerate(weights_for_equations):
                color = self.colors[i % len(self.colors)]
                self._plot_single(weights_for_equation, color=color)
        else:
            weights_for_equation = cast(List[float], weights_for_equations)
            self._plot_single(weights_for_equation, self.colors[0])
        plt.ylabel("Expected Return")
        plt.xlabel("Portfolio Efficiency")
        plt.title("Expected Return v.s. Portfolio Efficiency")
        plt.grid(True)
        plt.show()



from typing import TYPE_CHECKING, List

import numpy as np
import optuna

from ..evaluation import Calculator

if TYPE_CHECKING:
    from .multiple_objective import MultipleObjective


def construct_power_weights(
    ob: "MultipleObjective", trial: optuna.Trial
) -> List[float]:
    """
    Construct power weights based on the attributes of the MultipleObjective instance and the current trial.

    Args:
        ob (MultipleObjective): An instance of the MultipleObjective class, guiding the construction of the power weights.
        trial (optuna.Trial): The current optuna trial from which to suggest values for the weights.

    Returns:
        List[float]: A list of power weights constructed based on the given MultipleObjective instance and the current trial.
    """
    power_weights: List[float] = []
    if ob.weights_num is None:
        ob.weights_num = ob.get_weights_num()

    if ob.dirichlet:
        for i in range(ob.weights_num - 1):
            power_weights.append(
                trial.suggest_float(
                    f"w_po_{i+1}", 0, max((1 - sum(power_weights), 0.1))
                )
            )
        power_weights.append(1 - sum(power_weights))
    else:
        for i in range(ob.weights_num):
            power_weights.append(
                trial.suggest_float(
                    f"w{i+1}", ob.power_lower_bound, ob.power_upper_bound
                )
            )

    return power_weights


def construct_first_order_weights(
    ob: "MultipleObjective", trial: optuna.Trial
) -> List[float]:
    """
    Construct first order weights based on the attributes of the MultipleObjective instance and the current trial.

    Args:
        ob (MultipleObjective): An instance of the MultipleObjective class, guiding the construction of the first order weights.
        trial (optuna.Trial): The current optuna trial from which to suggest values for the weights.

    Returns:
        List[float]: A list of first order weights constructed based on the given MultipleObjective instance and the current trial.
    """
    first_order_weights: List[float] = []
    log = ob.first_order_lower_bound >= 0

    if ob.weights_num is None:
        ob.weights_num = ob.get_weights_num()

    first_order_scale_upper_bound = ob.first_order_scale_upper_bound
    first_order_scale_lower_bound = ob.first_order_scale_lower_bound

    if isinstance(ob.calculator, Calculator):
        value_scales = ob.calculator.value_scales
        for i in range(ob.weights_num):
            first_order_weights.append(
                trial.suggest_float(
                    f"w_fo_{i+1}",
                    np.power(10, value_scales[i] - first_order_scale_lower_bound),
                    np.power(10, value_scales[i] + first_order_scale_upper_bound),
                    log=False,
                )
            )
        if ob.max_min_scale_ratio is not None:
            scales_with_weights = [
                weight * np.power(10, -value_scale)
                for weight, value_scale in zip(first_order_weights, value_scales)
            ]
            scales_with_weights_ratio = max(scales_with_weights) / (
                min(scales_with_weights) + 1e-6
            )
            first_order_weights = [
                weight * ob.max_min_scale_ratio / scales_with_weights_ratio
                for weight in first_order_weights
            ]
    else:
        for i in range(ob.weights_num):
            first_order_weights.append(
                trial.suggest_float(
                    f"w_fo_{i+1}",
                    ob.first_order_lower_bound,
                    ob.first_order_upper_bound,
                    log=log,
                )
            )

    return first_order_weights


def construct_log_pca_weights(
    ob: "MultipleObjective", trial: optuna.Trial
) -> List[float]:
    """
    Constructs a list of weights for log PCA components based on the optimization trial.

    This function generates a list of floating-point numbers representing the weights
    for log PCA components. Each weight is suggested by the trial within the bounds
    defined in the MultipleObjective instance.

    """
    log_pca_weights: List[float] = []
    if ob.weights_num is None:
        ob.weights_num = ob.get_weights_num()
    for i in range(ob.weights_num):
        log_pca_weights.append(
            trial.suggest_float(
                f"w{i+1}", ob.pca_importance_lower_bound, ob.pca_importance_upper_bound
            )
        )
    return log_pca_weights


def construct_weights(ob: "MultipleObjective", trial: optuna.Trial) -> List[float]:
    """
    Construct weights by combining power and first order weights as required by the MultipleObjective instance.

    Args:
        ob (MultipleObjective): An instance of the MultipleObjective class, guiding the construction of the weights.
        trial (optuna.Trial): The current optuna trial from which to suggest values for the weights.

    Returns:
        List[float]: A list of weights constructed based on the given MultipleObjective instance and the current trial.
    """
    weights = []
    if ob.calculator.equation_type == "log_pca":
        weights = construct_log_pca_weights(ob, trial)
    elif not (ob.first_order):
        weights = construct_power_weights(ob, trial)
    elif ob.calculator.equation_type == "product" and ob.power:
        weights = construct_power_weights(ob, trial) + construct_first_order_weights(
            ob, trial
        )
    elif ob.calculator.equation_type == "sum":
        weights = construct_first_order_weights(ob, trial)
    elif ob.calculator.equation_type == "free_style":
        weights = construct_first_order_weights(ob, trial)
    ob.calculator.get_overall_score(
        weights_for_equation=weights,
    )

    return weights



from typing import TYPE_CHECKING, Optional, Tuple

if TYPE_CHECKING:
    from .calculator import Calculator

from .base_evaluator import evaluation_preprocessor


@evaluation_preprocessor
def calculate_distinct_count_portfolio_concentration(
    calculator: "Calculator",
    target_column: str,
    mask_column: Optional[str] = None,
    expected_coverage: Optional[float] = None,
) -> Tuple[float, float]:
    """
    Calculate the threshold and concentration of a portfolio based on a specified target column and expected coverage.

    This function takes a Calculator object, which is expected to contain a DataFrame. It computes the unique count of
    values in the target column and determines a threshold in 'overall_score' such that a certain proportion
    (expected_coverage) of unique values is covered. The concentration is calculated as the proportion of rows
    exceeding this threshold.

    Args:
        calculator: An instance of Calculator which contains a DataFrame.
        target_column: The name of the column in the DataFrame to analyze.
        mask_column: The name of the column in the DataFrame to mask the data. Defaults to None.
        expected_coverage: The expected proportion of unique values to be covered. Defaults to 0.95 if not provided.

    Returns:
        A tuple containing:
        - threshold (float): The minimum 'overall_score' to cover the expected proportion of unique values.
        - concentration (float): The proportion of rows in the DataFrame with 'overall_score' above the threshold.
    """
    if expected_coverage is None:
        expected_coverage = 0.95
    df = calculator.evaluated_dataframe
    total_ids = df[target_column].nunique()

    df_sorted = df.sort_values(by="overall_score", ascending=False).reset_index(
        drop=True
    )

    df_sorted["is_unique"] = ~df_sorted[target_column].duplicated()
    df_sorted["cumulative_coverage"] = df_sorted["is_unique"].cumsum() / total_ids
    df_filtered = df_sorted[df_sorted["cumulative_coverage"] > expected_coverage]
    if not df_filtered.empty:
        threshold_row = df_filtered.iloc[0]
        threshold = threshold_row["overall_score"]
        concentration = len(
            df_sorted[df_sorted["overall_score"] > threshold][target_column]
        ) / len(df_sorted)
    else:
        threshold = 1
        concentration = 1
    return threshold, concentration



from typing import TYPE_CHECKING, List

import numpy as np

if TYPE_CHECKING:
    from .calculator import Calculator


def merge_and_count(
    target_queue: np.ndarray,
    temporary_queue: np.ndarray,
    start_index: int,
    end_index: int,
    weight_type: str,
) -> float:
    """
    A helper function for the merge sort algorithm. It is used to merge two sorted
    subarrays and count the inverse pairs based on the provided weights type.

    :param target_queue: The original target scores to be sorted and merged.
    :param temp_queue: A temporary array to store merged results.
    :param left: The starting index of the portion to be merged.
    :param right: The ending index of the portion to be merged.
    :param weights_type: The type of weights used in calculating inverse pairs.
    :return: The computed inverse pairs value for the merged segment.
    """

    def merge_ranges(
        target_queue: np.ndarray,
        temporary_queue: np.ndarray,
        i: int,
        j: int,
        end: int,
        weight_type: str,
    ) -> float:
        """
        Helper function to merge two ranges and count inverse pairs based on the provided weights type.
        """

        result = 0.0
        batch_size, _ = target_queue.shape

        for k in range(batch_size):
            idx_temporary, idx_left, idx_right = i, i, j
            while idx_left < j and idx_right < end:
                if target_queue[k, idx_left] >= target_queue[k, idx_right]:
                    temporary_queue[k, idx_temporary] = target_queue[k, idx_left]
                    idx_left += 1
                else:
                    temporary_queue[k, idx_temporary] = target_queue[k, idx_right]
                    idx_right += 1
                    if weight_type == "count":
                        result += 1
                    elif weight_type == "linear":
                        result += (j - idx_left) * 0.5
                    elif weight_type == "exponential":
                        result += (j - idx_left) * (0.8**k)
                idx_temporary += 1

            temporary_queue[k, idx_temporary:j] = target_queue[k, idx_left:j]
            temporary_queue[k, idx_temporary:end] = target_queue[k, idx_right:end]

        return float(result)

    length = end_index - start_index
    step = 1
    result = 0.0

    while step < length:
        for i in range(start_index, end_index, 2 * step):
            j = min(i + step, end_index)
            end = min(j + step, end_index)
            result += merge_ranges(
                target_queue, temporary_queue, i, j, end, weight_type
            )
            target_queue[:, i:end] = temporary_queue[:, i:end]
        step *= 2

    return float(result)


def calculate_inverse_pairs(
    target_scores: np.ndarray, merge_scores: np.ndarray, weights_type: str
) -> float:
    """
    Calculate the inverse pairs for the given target scores based on merge scores.

    :param target_scores: The original target scores.
    :param merge_scores: The scores used to determine the order of merging.
    :param weights_type: The type of weights used in calculating inverse pairs.
    :return: The computed inverse pairs value.
    """

    num_samples, dimension = target_scores.shape
    sorted_indices = np.argsort(-merge_scores, axis=-1)
    sample_indices = np.arange(0, num_samples).reshape([-1, 1])
    sorted_target_scores = target_scores[sample_indices, sorted_indices]
    temp_scores = np.zeros((num_samples, dimension), dtype=np.float32)
    result = merge_and_count(
        sorted_target_scores, temp_scores, 0, dimension, weights_type
    )
    return float(result / num_samples)


def calculate_inverse_pair(
    calculator: "Calculator",
    weights_for_equation: List[float],
    weights_type: str = "count",
) -> float:
    """
    Calculates the weighted sum of inverse pairs for selected columns
    using the specified weighting scheme.

    :param weights_for_equation: The list of weights to be applied to each column.
    :param weights_type: The type of weights used for inverse pair calculation.
                            Supported values are "count", "linear", and "exponential".
    :return: The computed weighted sum of inverse pairs.
    """
    calculator.get_overall_score(weights_for_equation)

    result = 0.0
    for i, weight in enumerate(weights_for_equation):
        score = calculate_inverse_pairs(
            calculator.selected_values[:, i],
            calculator.df["overall_score"],
            weights_type,
        )
        result += score * weight
    return result



from collections import Counter
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from ..visualization.lorenz_curve import LorenzCurveGini
from .base import BaseSampler


class GiniSampler(BaseSampler):
    """
    This class provides methods to sample data according to Gini coefficient.
    """

    def __init__(
        self,
        sample_size: int,
        data: Union[pd.Series, List[float]],
        slice_from: Optional[float] = None,
        slice_to: Optional[float] = None,
        log_scale: Optional[bool] = True,
        laplace_smoothing: Optional[bool] = True,
        bounds_num: Optional[int] = None,
    ) -> None:
        """
        Initialize with a list of data.

        :param sample_size: number of samples
        :param data: a list of floats
        :param slice_from: lower bound of the data
        :param slice_to: upper bound of the data
        :param bounds_num: number of bounds
        :return: a list of floats
        """
        super().__init__(
            sample_size, data, slice_from, slice_to, log_scale, laplace_smoothing
        )
        self.lorenz_gini = LorenzCurveGini(self.data)
        if bounds_num is None:
            self.bounds_num = self.sample_size * 10
        else:
            self.bounds_num = bounds_num
        self.bounds = self.lorenz_gini.get_bounds(
            num_quantiles=self.bounds_num,
            slice_from=self.slice_from,
            slice_to=self.slice_to,
            unique_bounds=False,
        )

    def equidistant_indices(self, gini_list: List[float]) -> List[float]:
        """
        Get equidistant indices from Gini list.

        :param gini_list: a list of Gini coefficients
        """
        min_gini = np.min(gini_list)
        max_gini = np.max(gini_list)

        gini_segments = np.linspace(max_gini, min_gini, self.sample_size + 1)
        start_idx = 0

        segment_indices = []
        for g in gini_segments:
            segment_idx = start_idx + np.abs(gini_list[start_idx:] - g).argmin()
            segment_indices.append(segment_idx)
            start_idx = segment_idx

        segment_bounds = [self.bounds[i] for i in segment_indices]

        return segment_bounds

    def sample(self) -> dict:
        """
        Sample data according to Gini coefficient.
        """

        gini_list = self.lorenz_gini.get_gini_list_from_bounds(self.bounds)
        segment_bounds = self.equidistant_indices(gini_list)
        return dict(Counter(segment_bounds))



from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from .calculator import Calculator


def calculate_log_mse(calculator: "Calculator", target_column: str) -> float:
    """Calculate log mean squared error.

    :param target_column: target column
    """
    log_true = np.log(calculator.df[target_column] + 1)
    log_pred = np.log(calculator.df["overall_score"] + 1)
    mse = np.mean((log_true - log_pred) ** 2)
    return float(mse)
